@@ -13,7 +13,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
  *
  */
 
@@ -29,7 +29,7 @@
  * For each request, the parent checks if the dbus connection is still alive,
  * and pings the child to see if it's still responding.  If the child has
  * exited or is not responding, the parent starts a new metadata helper as
- * described above.  
+ * described above.
  *
  * The child process exits after a certain period of inactivity (30s
  * currently), so the ping message serves two purposes - it checks that the
@@ -65,11 +65,12 @@ static gboolean tried_env_address = FALSE;
 static DBusConnection *dbus_connection = NULL;
 static GPid metadata_child = 0;
 static GMainContext *main_context = NULL;
+static GStaticMutex conn_mutex = G_STATIC_MUTEX_INIT;
 
 struct RBMetaDataPrivate
 {
-	char *uri;
-	char *mimetype;
+	char       *uri;
+	char       *mimetype;
 	GHashTable *metadata;
 };
 
@@ -79,6 +80,7 @@ static void
 rb_metadata_class_init (RBMetaDataClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
 	object_class->finalize = rb_metadata_finalize;
 
 	g_type_class_add_private (object_class, sizeof (RBMetaDataPrivate));
@@ -98,6 +100,7 @@ rb_metadata_finalize (GObject *object)
 	RBMetaData *md;
 
 	md = RB_METADATA (object);
+
 	g_free (md->priv->uri);
 	g_free (md->priv->mimetype);
 	if (md->priv->metadata)
@@ -118,7 +121,11 @@ kill_metadata_service (void)
 	if (dbus_connection) {
 		if (dbus_connection_get_is_connected (dbus_connection)) {
 			rb_debug ("closing dbus connection");
+#ifdef WITH_OLD_DBUS
 			dbus_connection_disconnect (dbus_connection);
+#else
+			dbus_connection_close (dbus_connection);
+#endif
 		} else {
 			rb_debug ("dbus connection already closed");
 		}
@@ -150,9 +157,9 @@ ping_metadata_service (GError **error)
 	if (!message) {
 		return FALSE;
 	}
-	response = dbus_connection_send_with_reply_and_block (dbus_connection, 
-							      message, 
-							      RB_METADATA_DBUS_TIMEOUT, 
+	response = dbus_connection_send_with_reply_and_block (dbus_connection,
+							      message,
+							      RB_METADATA_DBUS_TIMEOUT,
 							      &dbus_error);
 	dbus_message_unref (message);
 	if (dbus_error_is_set (&dbus_error)) {
@@ -170,19 +177,6 @@ ping_metadata_service (GError **error)
 static gboolean
 start_metadata_service (GError **error)
 {
-	/*
-	 * Normally, we find the metadata helper in the libexec dir,
-	 * but when --enable-uninstalled-build is specified, we look
-	 * in the directory it's built in.
-	 */
-	char *argv[] = { 
-#ifdef METADATA_UNINSTALLED_DIR
-		METADATA_UNINSTALLED_DIR "/rhythmbox-metadata",
-#else
-		LIBEXEC_DIR "/rhythmbox-metadata", 
-#endif
-		"unix:tmpdir=/tmp", NULL 
-	};
 	DBusError dbus_error = {0,};
 	GIOChannel *stdout_channel;
 	GIOStatus status;
@@ -196,7 +190,7 @@ start_metadata_service (GError **error)
 		 * into any errors yet, we can try to restart it.
 		 */
 		kill_metadata_service ();
-		
+
 		if (*error)
 			return FALSE;
 	}
@@ -209,27 +203,56 @@ start_metadata_service (GError **error)
 			dbus_address = g_strdup (addr);
 			metadata_child = 0;
 		}
-	} 
+	}
 
 	if (dbus_address == NULL) {
 		gint metadata_stdout;
+		GPtrArray *argv;
+		gboolean res;
+		char **debug_args;
+		GError *local_error;
+		int i;
 
-		if (!g_spawn_async_with_pipes (NULL,
-					       argv,
-					       NULL,
-					       0,
-					       NULL, NULL,
-					       &metadata_child,
-					       NULL,
-					       &metadata_stdout,
-					       NULL,
-					       error)) {
+		argv = g_ptr_array_new ();
+		/*
+		 * Normally, we find the metadata helper in the libexec dir,
+		 * but when --enable-uninstalled-build is specified, we look
+		 * in the directory it's built in.
+		 */
+#ifdef METADATA_UNINSTALLED_DIR
+		g_ptr_array_add (argv, METADATA_UNINSTALLED_DIR "/rhythmbox-metadata");
+#else
+		g_ptr_array_add (argv, LIBEXEC_DIR G_DIR_SEPARATOR_S INSTALLED_METADATA_HELPER);
+#endif
+		debug_args = rb_debug_get_args ();
+		i = 0;
+		while (debug_args[i] != NULL) {
+			g_ptr_array_add (argv, debug_args[i]);
+			i++;
+		}
+
+		g_ptr_array_add (argv, "unix:tmpdir=/tmp");
+		g_ptr_array_add (argv, NULL);
+
+		local_error = NULL;
+		res = g_spawn_async_with_pipes (NULL,
+						(char **)argv->pdata,
+						NULL,
+						0,
+						NULL, NULL,
+						&metadata_child,
+						NULL,
+						&metadata_stdout,
+						NULL,
+						&local_error);
+		g_ptr_array_free (argv, TRUE);
+		g_strfreev (debug_args);
+
+		if (res == FALSE) {
+			g_propagate_error (error, local_error);
 			return FALSE;
 		}
 
-		/* hmm, probably shouldn't do this */
-		signal (SIGPIPE, SIG_IGN);
-		
 		stdout_channel = g_io_channel_unix_new (metadata_stdout);
 		status = g_io_channel_read_line (stdout_channel, &dbus_address, NULL, NULL, error);
 		g_io_channel_unref (stdout_channel);
@@ -254,7 +277,7 @@ start_metadata_service (GError **error)
 	dbus_connection_set_exit_on_disconnect (dbus_connection, FALSE);
 
 	dbus_connection_setup_with_g_main (dbus_connection, main_context);
-	
+
 	rb_debug ("Metadata process %d started", metadata_child);
 	return TRUE;
 }
@@ -306,11 +329,15 @@ rb_metadata_load (RBMetaData *md,
 		  const char *uri,
 		  GError **error)
 {
-	DBusMessage *message;
-	DBusMessage *response;
+	DBusMessage *message = NULL;
+	DBusMessage *response = NULL;
 	DBusMessageIter iter;
 	DBusError dbus_error = {0,};
 	gboolean ok;
+	GError *fake_error = NULL;
+
+	if (error == NULL)
+		error = &fake_error;
 
 	g_free (md->priv->mimetype);
 	md->priv->mimetype = NULL;
@@ -322,80 +349,83 @@ rb_metadata_load (RBMetaData *md,
 
 	if (md->priv->metadata)
 		g_hash_table_destroy (md->priv->metadata);
-	md->priv->metadata = g_hash_table_new (g_direct_hash, g_direct_equal);
+	md->priv->metadata = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)rb_value_free);
 
-	
-	if (!start_metadata_service (error))
-		return;
+	g_static_mutex_lock (&conn_mutex);
 
-	message = dbus_message_new_method_call (RB_METADATA_DBUS_NAME,
-						RB_METADATA_DBUS_OBJECT_PATH,
-						RB_METADATA_DBUS_INTERFACE,
-						"load");
-	if (!message) {
-		g_set_error (error,
-			     RB_METADATA_ERROR,
-			     RB_METADATA_ERROR_INTERNAL,
-			     _("D-BUS communication error"));
-		return;
-	}
-	if (!dbus_message_append_args (message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_INVALID)) {
-		g_set_error (error,
-			     RB_METADATA_ERROR,
-			     RB_METADATA_ERROR_INTERNAL,
-			     _("D-BUS communication error"));
-		return;
-	}
+	start_metadata_service (error);
 
-	rb_debug ("sending metadata load request");
-	response = dbus_connection_send_with_reply_and_block (dbus_connection, 
-							      message, 
-							      RB_METADATA_DBUS_TIMEOUT, 
-							      &dbus_error);
-
-	dbus_message_unref (message);
-	if (!response) {
-		handle_dbus_error (md, &dbus_error, error);
-		return;
-	}
-
-	if (!dbus_message_iter_init (response, &iter)) {
-		g_set_error (error,
-			     RB_METADATA_ERROR,
-			     RB_METADATA_ERROR_INTERNAL,
-			     _("D-BUS communication error"));
-		rb_debug ("couldn't read response message");
-		dbus_message_unref (response);
-		return;
-	}
-
-	if (!rb_metadata_dbus_get_boolean (&iter, &ok)) {
-		g_set_error (error,
-			     RB_METADATA_ERROR,
-			     RB_METADATA_ERROR_INTERNAL,
-			     _("D-BUS communication error"));
-		rb_debug ("couldn't get success flag from response message");
-		dbus_message_unref (response);
-		return;
-	}
-
-	if (ok) {
-		/* get mime type */
-		if (!rb_metadata_dbus_get_string (&iter, &md->priv->mimetype)) {
+	if (*error == NULL) {
+		message = dbus_message_new_method_call (RB_METADATA_DBUS_NAME,
+							RB_METADATA_DBUS_OBJECT_PATH,
+							RB_METADATA_DBUS_INTERFACE,
+							"load");
+		if (!message) {
+			g_set_error (error,
+				     RB_METADATA_ERROR,
+				     RB_METADATA_ERROR_INTERNAL,
+				     _("D-BUS communication error"));
+		} else if (!dbus_message_append_args (message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_INVALID)) {
 			g_set_error (error,
 				     RB_METADATA_ERROR,
 				     RB_METADATA_ERROR_INTERNAL,
 				     _("D-BUS communication error"));
-		} else {
-			/* get metadata */
-			rb_debug ("got mimetype: %s", md->priv->mimetype);
-			rb_metadata_dbus_read_from_message (md, md->priv->metadata, &iter);
 		}
-	} else {
-		read_error_from_message (md, &iter, error);
 	}
 
-	dbus_message_unref (response);
+	if (*error == NULL) {
+		rb_debug ("sending metadata load request");
+		response = dbus_connection_send_with_reply_and_block (dbus_connection,
+								      message,
+								      RB_METADATA_DBUS_TIMEOUT,
+								      &dbus_error);
+
+		if (!response)
+			handle_dbus_error (md, &dbus_error, error);
+	}
+
+	if (*error == NULL) {
+		if (!dbus_message_iter_init (response, &iter)) {
+			g_set_error (error,
+				     RB_METADATA_ERROR,
+				     RB_METADATA_ERROR_INTERNAL,
+				     _("D-BUS communication error"));
+			rb_debug ("couldn't read response message");
+		}
+	}
+
+	if (*error == NULL) {
+		if (!rb_metadata_dbus_get_boolean (&iter, &ok)) {
+			g_set_error (error,
+				     RB_METADATA_ERROR,
+				     RB_METADATA_ERROR_INTERNAL,
+				     _("D-BUS communication error"));
+			rb_debug ("couldn't get success flag from response message");
+		} else if (ok) {
+			/* get mime type */
+			if (!rb_metadata_dbus_get_string (&iter, &md->priv->mimetype)) {
+				g_set_error (error,
+					     RB_METADATA_ERROR,
+					     RB_METADATA_ERROR_INTERNAL,
+					     _("D-BUS communication error"));
+			} else {
+				/* get metadata */
+				rb_debug ("got mimetype: %s", md->priv->mimetype);
+				rb_metadata_dbus_read_from_message (md, md->priv->metadata, &iter);
+			}
+		} else {
+			read_error_from_message (md, &iter, error);
+		}
+	}
+
+	if (message)
+		dbus_message_unref (message);
+	if (response)
+		dbus_message_unref (response);
+	if (fake_error)
+		g_error_free (fake_error);
+
+	g_static_mutex_unlock (&conn_mutex);
 }
 
 const char *
@@ -427,7 +457,7 @@ rb_metadata_set (RBMetaData *md, RBMetaDataField field,
 {
 	GValue *newval;
 	GType type;
-	
+
 	type = rb_metadata_get_field_type (field);
 	g_return_val_if_fail (type == G_VALUE_TYPE (val), FALSE);
 
@@ -444,90 +474,112 @@ gboolean
 rb_metadata_can_save (RBMetaData *md, const char *mimetype)
 {
 	GError *error = NULL;
-	DBusMessage *message;
-	DBusMessage *response;
+	DBusMessage *message = NULL;
+	DBusMessage *response = NULL;
 	gboolean can_save = FALSE;
 	DBusError dbus_error = {0,};
 	DBusMessageIter iter;
+	gboolean ok = TRUE;
+
+	g_static_mutex_lock (&conn_mutex);
 
 	if (start_metadata_service (&error) == FALSE) {
 		g_error_free (error);
-		return FALSE;
+		ok = FALSE;
 	}
 
-	message = dbus_message_new_method_call (RB_METADATA_DBUS_NAME,
-						RB_METADATA_DBUS_OBJECT_PATH,
-						RB_METADATA_DBUS_INTERFACE,
-						"canSave");
-	if (!message) {
-		return FALSE;
-	}
-	if (!dbus_message_append_args (message, DBUS_TYPE_STRING, &mimetype, DBUS_TYPE_INVALID)) {
-		return FALSE;
+	if (ok) {
+		message = dbus_message_new_method_call (RB_METADATA_DBUS_NAME,
+							RB_METADATA_DBUS_OBJECT_PATH,
+							RB_METADATA_DBUS_INTERFACE,
+							"canSave");
+		if (!message) {
+			ok = FALSE;
+		} else if (!dbus_message_append_args (message, DBUS_TYPE_STRING, &mimetype, DBUS_TYPE_INVALID)) {
+			ok = FALSE;
+		}
 	}
 
-	response = dbus_connection_send_with_reply_and_block (dbus_connection, 
-							      message, 
-							      RB_METADATA_DBUS_TIMEOUT, 
-							      &dbus_error);
-	dbus_message_unref (message);
-	if (!response) {
-		dbus_error_free (&dbus_error);
-		return FALSE;
+	if (ok) {
+		response = dbus_connection_send_with_reply_and_block (dbus_connection,
+								      message,
+								      RB_METADATA_DBUS_TIMEOUT,
+								      &dbus_error);
+		if (!response) {
+			dbus_error_free (&dbus_error);
+			ok = FALSE;
+		} else if (dbus_message_iter_init (response, &iter)) {
+			rb_metadata_dbus_get_boolean (&iter, &can_save);
+		}
 	}
 
-	if (dbus_message_iter_init (response, &iter)) {
-		rb_metadata_dbus_get_boolean (&iter, &can_save);
-	}
-	dbus_message_unref (response);
+	if (message)
+		dbus_message_unref (message);
+	if (response)
+		dbus_message_unref (response);
+	g_static_mutex_unlock (&conn_mutex);
+
 	return can_save;
 }
 
 void
 rb_metadata_save (RBMetaData *md, GError **error)
 {
-	DBusMessage *message;
-	DBusMessage *response;
+	GError *fake_error = NULL;
+	DBusMessage *message = NULL;
+	DBusMessage *response = NULL;
 	DBusError dbus_error = {0,};
 	DBusMessageIter iter;
 
-	if (start_metadata_service (error) == FALSE)
-		return;
+	if (error == NULL)
+		error = &fake_error;
 
-	message = dbus_message_new_method_call (RB_METADATA_DBUS_NAME,
-						RB_METADATA_DBUS_OBJECT_PATH,
-						RB_METADATA_DBUS_INTERFACE,
-						"save");
-	if (!message) {
-		g_set_error (error,
-			     RB_METADATA_ERROR,
-			     RB_METADATA_ERROR_INTERNAL,
-			     _("D-BUS communication error"));
-		return;
+	g_static_mutex_lock (&conn_mutex);
+
+	start_metadata_service (error);
+
+	if (*error == NULL) {
+		message = dbus_message_new_method_call (RB_METADATA_DBUS_NAME,
+							RB_METADATA_DBUS_OBJECT_PATH,
+							RB_METADATA_DBUS_INTERFACE,
+							"save");
+		if (!message) {
+			g_set_error (error,
+				     RB_METADATA_ERROR,
+				     RB_METADATA_ERROR_INTERNAL,
+				     _("D-BUS communication error"));
+		}
 	}
 
-	dbus_message_iter_init_append (message, &iter);
-	if (!rb_metadata_dbus_add_to_message (md, &iter)) {
-		g_set_error (error,
-			     RB_METADATA_ERROR,
-			     RB_METADATA_ERROR_INTERNAL,
-			     _("D-BUS communication error"));
-		return;
+	if (*error == NULL) {
+		dbus_message_iter_init_append (message, &iter);
+		if (!rb_metadata_dbus_add_to_message (md, &iter)) {
+			g_set_error (error,
+				     RB_METADATA_ERROR,
+				     RB_METADATA_ERROR_INTERNAL,
+				     _("D-BUS communication error"));
+		}
 	}
 
-	response = dbus_connection_send_with_reply_and_block (dbus_connection, 
-							      message, 
-							      RB_METADATA_DBUS_TIMEOUT, 
-							      &dbus_error);
-	dbus_message_unref (message);
-	if (!response) {
-		handle_dbus_error (md, &dbus_error, error);
-		return;
+	if (*error == NULL) {
+		response = dbus_connection_send_with_reply_and_block (dbus_connection,
+								      message,
+								      RB_METADATA_DBUS_TIMEOUT,
+								      &dbus_error);
+		if (!response) {
+			handle_dbus_error (md, &dbus_error, error);
+		} else if (dbus_message_iter_init (response, &iter)) {
+			/* if there's any return data at all, it'll be an error */
+			read_error_from_message (md, &iter, error);
+		}
 	}
 
-	/* if there's any return data at all, it'll be an error */
-	if (dbus_message_iter_init (response, &iter)) {
-		read_error_from_message (md, &iter, error);
-	}
+	if (message)
+		dbus_message_unref (message);
+	if (response)
+		dbus_message_unref (response);
+	if (fake_error)
+		g_error_free (fake_error);
+
+	g_static_mutex_unlock (&conn_mutex);
 }
-
